/*! \file can.c \brief CAN driver with limited can support. */
// *****************************************************************************
//
// File Name	: 'can.c'
// Title		: CAN driver with limited can support
// Author		: Markus Schiesser 2007
// Created		: 01/01/2007
// Revised		: 04/27/2007
// Version		: 0.7
// Target MCU	: ATMEL AVR Series
// Editor Tabs	: 3
//
// Supports:	extended frame ID (CAN 2.0B)
//				interrupt driven RX/TX
//				polling RX/TX
//				up to 15 message objects
// Not supported:
//				standart frame ID (CAN2.0A)
//				remote frames
//				Frame buffer mode
//				and few other things		
//
// *****************************************************************************


#include "config.h"
#include "library/can.h"
#include "library/can_config.h"


extern struct MOb_struct MOb_TX;
extern struct MOb_struct MOb_RX;
extern struct CAN_Error_struct CAN_Counter;



//CAN init
void CAN_Init()
{
	uint8_t mob;
	
	// clear up all MOb registers
	for(mob = 0; mob < MAX_MOB; mob++)	{
		CAN_Clear_MOb(mob);
	 }

	// set CAN baudrate (values predefined in can.h)
	CANBT1 = CONF_CANBT1;
	CANBT2 = CONF_CANBT2;
	CANBT3 = CONF_CANBT3;
	
	// set CAN interruptes predefined in can_config.h)
	CAN_Set_IRQ();							
	
	//enable CAN subsystem
	CAN_ENABLE;							
}

void CAN_Set_IRQ()
{
	// Enable / Disable bus off interrupte
	#ifndef CAN_BOFF_INTERRUPT
	#	error CAN_BOFF_INTERRUPT not defind in config.h
	#endif

	#if (CAN_BOFF_INTERRUPT == ENABLE)
		CAN_ENBOFF_IRQ_ENABLE;	
	#else 
		CAN_ENBOFF_IRQ_DISABLE;
	#endif

	// Enable / Disable receive interrupt
	#ifndef CAN_RXOK_INTERRUPT
	#	error CAN_RXOK_INTERRUPT not defined in can_config.h
	#endif

	#if (CAN_RXOK_INTERRUPT == ENABLE)
		CAN_RX_IRQ_ENABLE;		
	#else 
		CAN_RX_IRQ_DISABLE;
	#endif

	// Enable / Disable transmit interrupt
	#ifndef CAN_TXOK_INTERRUPT
	#	error CAN_TXOK_INTERRUPT not defined in can_config.h
	#endif

	#if CAN_TXOK_INTERRUPT == ENABLE
		CAN_TX_IRQ_ENABLE;		
	#else
		CAN_TX_IRQ_DISABLE;
	#endif

	// Enable / Disable MOb error interrupt
	#ifndef CAN_ERROR_MOB_INTERRUPT
	#	error CAN_ERROR_MOB_INTERRUPT not defined in can_config.h
	#endif

	#if CAN_ERROR_MOB_INTERRUPT == ENABLE
		CAN_ENERR_IRQ_ENABLE;	
	#else
		CAN_ENERR_IRQ_DISABLE;
	#endif

	// Enable / Disable frame buffer interrupt
	#ifndef CAN_FRAMEBUF_INTERRUPT
	#	error CAN_FRAMEBUF_INTERRUPT not defined in can_config.h
	#endif

	#if CAN_FRAMBUF_INTERRUPT == ENABLE
		CAN_ENBX_IRQ_ENABLE;	
	#else
		CAN_ENBX_IRQ_DISABLE;
	#endif

	// Enable / Disable general error interrupt
	#ifndef CAN_GEN_ERROR_INTERRUPT
	#	error CAN_GEN_ERROR_INTERRUPT not defined in can_config.h
	#endif

	#if CAN_GEN_ERROR_INTERRUPT == ENABLE
		CAN_ENERG_IRQ_ENABLE;	
	#else
		CAN_ENERG_IRQ_DISABLE;
	#endif

	// Enable / Disable timer overrun interrupt
	#ifndef CAN_TIMER_INTERRUPT
	#	error CAN_TIMER_INTERRUPT not defined in can_config.h
	#endif

	#if CAN_TIMER_INTERRUPT == ENABLE
		CAN_ENOVRT_IRQ_ENABLE;	
	#else	
		CAN_ENOVRT_IRQ_DISABLE;	
	#endif

	// Enable / Disable general CAN interrupt
	#ifndef CAN_INTERRUPT
	#	error CAN_INTERRUPT not defined in can_config.h
	#endif
	
	#if CAN_INTERRUPT == ENABLE	
		CAN_IRQ_ENABLE;		
	#else
		CAN_IRQ_DISABLE;	
	#endif
}

// Clean all register concering mob 
void CAN_Clear_MOb(uint8_t mob)
{
	uint8_t save_canpage;
	
	//save current CANPAGE
	save_canpage = CANPAGE;
		
	//set CANPAGE
    CAN_SET_CANPAGE(mob);  
	//Reset CAN MOb Status Register
    CANSTMOB =0;
	//Reset CAN Mob Control and DLC Register
	CANCDMOB =0;
	//Reset CAN Time Stamp low Register 
	CANSTML  =0;					
	//Reset CAN Time Stamp high Register
	CANSTMH  =0;					
	//restore CANPAGE
	CANPAGE = save_canpage;
} 

//checks whether the mob is free
uint8_t CAN_Get_Free_MOb(uint8_t mob)
{
	uint8_t save_canpage;

	//save current CANPAGE
	save_canpage = CANPAGE;		       
	
	//set CANPAGE
	CAN_SET_CANPAGE(mob);

	if ((CANCDMOB & 0xC0) != 0x00)
		{
			// MOb is not free and restore CANPAGE
			CANPAGE = save_canpage;
			return FALSE; 			
		}

	//MOb is free and restore CANPAGE
	CANPAGE = save_canpage;
	return TRUE;					
}	

//Get MOb with interrupt
uint8_t CAN_Find_MOb_IRQ(void)
{
	uint8_t mob;
	
	mob = (CANHPMOB >> 4);
	
	return ((mob <= MAX_MOB) ? mob : FALSE);
}


//Configure the message object as RX
uint8_t CAN_SetRXMOb(uint8_t mob, uint32_t CAN_ID_Low, uint32_t CAN_ID_High, uint8_t interrupt)
{
	uint8_t save_canpage;
	uint32_t mask_ext,canID;

	if (mob > MAX_MOB) {
			//mob_number > max MOb
			return FALSE;			
		}
	
	if (CAN_Get_Free_MOb(mob) == FALSE) {
			//MOb not free
			return FALSE;			
		}

	//save current CANPAGE
	save_canpage = CANPAGE;

	//set new CANPAGE		       
	CAN_SET_CANPAGE(mob);

	//calculate and set mask
	mask_ext = ~(CAN_ID_High - CAN_ID_Low);
	CAN_SET_EXT_MASK(mask_ext);

	//calculate and set CAN ID
	canID = (mask_ext & CAN_ID_Low);
	CAN_SET_EXT_ID(canID);

	//enable / disable interrupt
	if (interrupt == ENABLE){
		//enable interrupt
		CAN_IRQ_ENABLE_MOB(mob);
	}else{
		CAN_IRQ_DISABLE_MOB(mob);
	}
	//activate MOb
	ENABLE_CHANNEL_RX;		

	//restore CANPAGE
	CANPAGE = save_canpage;                   

	return TRUE;
}	

//Configure the message object as TX
uint8_t CAN_SetTXMOb(uint8_t mob, uint8_t interrupt)
{
	uint8_t save_canpage;

	if (mob > MAX_MOB) {
			//mob > max MOb
			return FALSE;		
		}
	
	if (CAN_Get_Free_MOb(mob) == FALSE) {
			//MOb not free
			return FALSE;			
		}

	//save current CANPAGE
	save_canpage = CANPAGE;

	//set new CANPAGE		       
	CAN_SET_CANPAGE(mob);

		
    //configure MOb as CAN Rev 2.0B (extended)
	CANCDMOB |= (1<<IDE);

	// swicht interrupt on - off
	if (interrupt == ENABLE){
		//interrupt enabled
		CAN_IRQ_ENABLE_MOB(mob);
	}else{
		//interrupt disabled
		CAN_IRQ_DISABLE_MOB(mob);
	}

	//restore CANPAGE
	CANPAGE = save_canpage; 

	return TRUE;
}

// Transmitt CAN Frame
uint8_t CAN_TX_Frame(struct MOb_struct *pt_mob)
{
	uint8_t i,save_canpage;

	//check if MOb_struct_TX has valid datas
	if (pt_mob->valid == FALSE)			
		{
			// if not error and drop frame
			CAN_Counter.TX_Frame_lost++; 	
			return FALSE;
		}

	//save current CANPAGE
	save_canpage = CANPAGE;		       
	
	//set new CANPAGE
	CAN_SET_CANPAGE(pt_mob->MOb_number);

	//clean up all pending opterations
	CANSTMOB = 0x00;    	
	CANCDMOB = 0x00;
	
	//put address
	CAN_SET_EXT_ID(pt_mob->ext_id);	
		
	//put Data Length Code DLC	
	CANCDMOB|=pt_mob->data_length;		
	
	for (i=0; i < pt_mob->data_length; i++) {
			//store datas into CANMSG
			CANMSG=pt_mob->msg[i];	
			//count transmitted bytes		
			CAN_Counter.Byte_TX++;	
		}
	//count transmitted frames
	CAN_Counter.TXOK_count++;

	//enable transmission
	ENABLE_CHANNEL_TX; 

	//Resoter CANPAGE
	CANPAGE = save_canpage;                     
	
	return TRUE;
}


uint8_t CAN_TX_OK_Poll(struct MOb_struct *pt_mob)
{
	uint8_t save_canpage;
	
	//store current CANPAGE
	save_canpage=CANPAGE;
	
	//set CANPAGE
	CAN_SET_CANPAGE(pt_mob->MOb_number);
	
	if ((CANGSTA & (1<<TXBSY)) != FALSE) {
		//CAN is busy with transmission still
		//restore CANPAGE
		CANPAGE=save_canpage;
		return FALSE;
	}else{
		//transmission is finished
		pt_mob->valid = FALSE;
	}

	//restore CANPAGE
	CANPAGE=save_canpage;

	return TRUE;
}	
	
uint8_t CAN_RX_OK_Poll(struct MOb_struct *pt_mob)
{
	uint8_t save_canpage;
	
	//store current CANPAGE
	save_canpage=CANPAGE;
	
	//set CANPAGE
	CAN_SET_CANPAGE(pt_mob->MOb_number);
	
	if ((CANGSTA &(1<<RXOK)) == FALSE){
		//CAN is busy with receiving
		//restore CANPAGE
		CANPAGE=save_canpage;
		return FALSE;
	}else{	
		//all datas received 
		pt_mob->valid = TRUE;
		
	}
	//restore CANPAGE
	CANPAGE=save_canpage;

	return TRUE;
}		


//Called by interrupt
void CAN_Fct_RXOK_IRQ(uint8_t mob,struct MOb_struct *pt_mob)
{
	uint8_t i, save_canpage;
	
	//check if MOb_struct is free
	if (pt_mob->valid == TRUE){
			//if not error and drop frame
			CAN_Counter.RX_Frame_lost++; 
			//Reset RXOK interrupt
			CAN_RXOK_IRQ_CLEAR;	
			return;
		}

	//save current CANPAGE
	save_canpage = CANPAGE;
		
	//set CANPAGE
    CAN_SET_CANPAGE(mob); 

	//save CAN ID
	CAN_GET_EXT_ID(pt_mob->ext_id);
			
	//save MOb ID
	pt_mob->MOb_number = CAN_GET_CANPAGE;
			
	//save Data length DLC
	pt_mob->data_length = GET_DATA_LENGTH_CODE;

	//save Data into the struct and update
	for (i = 0; i < pt_mob->data_length; ++i){
			//read receide datas and store in struct
			pt_mob->msg[i]=CANMSG;

			//count received bytes	
			CAN_Counter.Byte_RX++;	
		}
	//count received frames
	CAN_Counter.RXOK_count++;

	//mark RX frame as valid	
	pt_mob->valid = TRUE;
					
	//Reset RXOK interrupt
	CAN_RXOK_IRQ_CLEAR;	
	
	//restore CANPAGE
	CANPAGE = save_canpage;		
}

//Called by interrupt
void CAN_Fct_TXOK_IRQ(uint8_t mob)
{	
	//mark TX frame as valid
	MOb_TX.valid = FALSE;
	
	//Reset TXOK interrupt				
	CAN_TXOK_IRQ_CLEAR;
}

void CAN_Fct_Error_IRQ(uint8_t mob)
{
	uint8_t save_canpage;

	//save current CANPAGE
	save_canpage = CANPAGE;
		
	//set CANPAGE
    CAN_SET_CANPAGE(mob); 
	
	// if Data Length Code warning and store error
	if ((CANSTMOB & (1 << DLCW)) != FALSE) CAN_Counter.DLCW_count++;
	// if Bit Error and store error
	if ((CANSTMOB & (1 << BERR)) != FALSE) CAN_Counter.BERR_count++;
	// if Stuff Error and store error
	if ((CANSTMOB & (1 << SERR)) != FALSE) CAN_Counter.SERR_count++;
	// if CRC Error and store error
	if ((CANSTMOB & (1 << CERR)) != FALSE) CAN_Counter.CRC_count++;
	// if Form Error and store error
	if ((CANSTMOB & (1 << FERR)) != FALSE) CAN_Counter.FERR_count++;
	// if Acknowledgment Error and store error
	if ((CANSTMOB & (1 << AERR)) != FALSE) CAN_Counter.AERR_count++;

	//Reset CAN Error interrupt
	CAN_ERROR_IRQ_CLEAR;
	
	//restor old CANPAGE
	CANPAGE = save_canpage;	
}

ISR (CANIT_vect)
{
	uint8_t save_canpage, mob;

	//save current page
	save_canpage = CANPAGE;	
		      
	//get firts mob with interrupt
	mob = CAN_Find_MOb_IRQ();
	
	if (mob != MAX_MOB) {
		CAN_SET_CANPAGE(mob);

		// if RX interrupt is present
		if ((CANSTMOB & (1<<RXOK)) != FALSE) {
			CAN_Fct_RXOK_IRQ(mob,&MOb_RX);
		}
		// if TX interrupt is present
		if((CANSTMOB & (1<<TXOK)) != FALSE) {
			CAN_Fct_TXOK_IRQ(mob);
		}
		// if ERROR interrupt is present
		if((CANSTMOB & ((1<<TXOK) | (1<<RXOK))) != FALSE) {
			CAN_Fct_Error_IRQ(mob);
		}

	}else{
		// no MOb matches with the interrupt => General IT
		CAN_GENERAL_IRQ_CLEAR;
	}
	CANPAGE = save_canpage;
}

/* CAN Timer Overrun */
ISR (OVRIT_vect)	
{
//not supported yet 

}



